/*
  Copyright 2004 Paul Zimmermann.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#define LLL_A 1
#define LLL_B 4

/* type of Gram-Schmidt coefficients */
#define MPZ_TYPE 0
#define DBL_TYPE 1
#define MPFR_TYPE 2
#define LDBL_TYPE 3
#define MPF_TYPE 4

#define TYPE MPF_TYPE /* type of Gram-Schmidt coefficients */
#define PRECISION 192  /* precision of Gram-Schmidt coefficients */
/* #define CUT 102 */

#ifdef DEBUG
#define CHECK(a) if (mpz_sizeinbase(a,2)>max_size) { \
                     max_size = mpz_sizeinbase(a,2); \
                     printf ("coeff size=%u\n", max_size); }
#else
#define CHECK(a)
#endif

#if (TYPE == MPZ_TYPE)
#define GS_FRACTION_FREE
#define GSTYPE mpz_t
#define GS_STRING "mpz"
#define INIT mpz_init
#define GS_MUL(a,b,c) { mpz_mul(a,b,c); CHECK(a); }
#define GS_SUBMUL(a,b,c,_t) { mpz_submul(a,b,c); CHECK(a); }
#define GS_DIVEXACT mpz_divexact
#define GS_SET mpz_set
#define GS_SET_Z mpz_set
#define GS_ISZERO(x) mpz_cmp_ui (x, 0) == 0
#define GS_ADDMUL(a,b,c,_t) { mpz_addmul(a,b,c); CHECK(a); }
#define GS_MUL_UI(a,b,c) { mpz_mul_ui(a,b,c); CHECK(a); }
#define GS_MUL_2EXP(a,b,c) { mpz_mul_2exp(a,b,c); CHECK(a); }
#if (LLL_A == 1)
#define GS_MUL_A(a,b)
#else
#define GS_MUL_A(a,b) mpz_mul_ui(a,b,LLL_A)
#endif
#if (LLL_B == 4)
#define GS_MUL_B(a,b) mpz_mul_2exp(a,b,2)
#define GS_DIV_B(a,b) mpz_tdiv_q_2exp(a,b,2)
#else
#define GS_MUL_B(a,b) mpz_mul_ui(a,b,LLL_B)
#define GS_DIV_B(a,b) mpz_tdiv_q_ui(a,b,LLL_B)
#endif
#define GS_DIV_UI mpz_tdiv_q_ui
#define GS_DIV_2EXP mpz_tdiv_q_2exp
#define GS_LT(a,b) mpz_cmp (a, b) < 0
#define CLEAR mpz_clear
#define GS_GTABS(a,b) mpz_cmpabs(a, b) > 0
#define GS_SWAP mpz_swap
/* r <- round(a/(2b)) = floor ((a+b)/2/b) */
#define GS_ROUND_DIV(r,a,b) mpz_add (r, a, b);         \
                            mpz_tdiv_q_2exp (r, r, 1); \
                            mpz_fdiv_q (r, r, b);
#define MPZ_SET_GS mpz_set
#define GS_OUT(x) mpz_out_str (stdout, 10, x); printf ("\n")
#elif (TYPE == DBL_TYPE)
#define GSTYPE double
#define GS_STRING "double"
#define GS_HAS_PREC 0
#define INIT(x) x = 0.0
#define GS_SUB(a,b,c) a = b - c
#define GS_MUL(a,b,c) a = b * c
#define GS_SUBMUL(a,b,c,_t) a -= b * c
#define GS_DIVEXACT(a,b,c) a = b / c
#define GS_SET(a,b) a = b
#define GS_SET_Z(a,b) a = mpz_get_d (b)
#define GS_ISZERO(x) x == 0.0
#define GS_ADDMUL(a,b,c,_t) a += b * c
#define GS_MUL_UI(a,b,c) a = b * (double) c
#define GS_DIV_UI(a,b,c) a = b / (double) c
#define GS_DIV(a,b,c) a = b / c
#define GS_LT(a,b) a < b
#define GS_LT_UI(a,b) a < (double) b
#define CLEAR(x)
#define GS_GTABS(a,b) (a > b) || (-a > b)
#define GS_GTABS_ONEHALF(x) ((x > 0.5) || (x < -0.5))
#define GS_SWAP(a, b) { double t; t = a; a = b; b = t; }
#define GS_ROUND_DIV(r,a,b) r = round (a / (2.0 * b))
#define GS_ROUND(r,a) r = round (a)
#define MPZ_SET_GS mpz_set_d
#define GS_OUT(x) printf ("%e\n", x)
#elif (TYPE == LDBL_TYPE)
#define GSTYPE long double
#define GS_STRING "long double"
#define GS_HAS_PREC 0
#define INIT(x) x = (long double) 0.0
#define GS_MUL(a,b,c) a = b * c
#define GS_SUBMUL(a,b,c,_t) a -= b * c
#define GS_DIVEXACT(a,b,c) a = b / c
#define GS_SET(a,b) a = b
#define GS_SET_Z(a,b) a = (long double) mpz_get_d (b)
#define GS_ISZERO(x) x == (long double) 0.0
#define GS_ADDMUL(a,b,c,_t) a += b * c
#define GS_MUL_UI(a,b,c) a = b * (long double) c
#define GS_DIV_UI(a,b,c) a = b / (long double) c
#define GS_LT(a,b) a < b
#define CLEAR(x)
#define GS_GTABS(a,b) (a > b) || (-a > b)
#define GS_SWAP(a, b) { long double t; t = a; a = b; b = t; }
#define GS_ROUND_DIV(r,a,b) r = (long double) round ((double) (a / ((long double) 2.0 * b)))
#define MPZ_SET_GS(a,b) mpz_set_d(a, (long double) b)
#define GS_OUT(x) printf ("%le\n", x)
#elif (TYPE == MPFR_TYPE)
#include "mpfr.h"
void wclr_mpz_set_fr (mpz_t, mpfr_t);
#define GSTYPE mpfr_t
#define GS_STRING "mpfr"
#define GS_HAS_PREC 1
#define GS_SET_DEFAULT_PREC mpfr_set_default_prec
#define GS_GET_PREC mpfr_get_prec
#define GS_SET_PREC(a,p) mpfr_prec_round(a,p,GMP_RNDN)
#define INIT(x) mpfr_init2 (x, PRECISION)
#define GS_MUL(a,b,c) mpfr_mul(a, b, c, GMP_RNDN)
#define GS_SUBMUL(a,b,c,_t) { mpfr_mul (_t, b, c, GMP_RNDN); \
                              mpfr_sub (a, a, _t, GMP_RNDN); }
#define GS_DIVEXACT(a,b,c) mpfr_div (a, b, c, GMP_RNDN)
#define GS_SET(a,b) mpfr_set (a, b, GMP_RNDN)
#define GS_SET_Z(a,b) mpfr_set_z (a, b, GMP_RNDN)
#define GS_ISZERO(x) mpfr_cmp_ui (x, 0) == 0
#define GS_SUB(a,b,c) mpfr_sub(a,b,c,GMP_RNDN)
#define GS_ADDMUL(a,b,c,_t) { mpfr_mul (_t, b, c, GMP_RNDN); \
                             mpfr_add (a, a, _t, GMP_RNDN); }
#define GS_MUL_UI(a,b,c) mpfr_mul_ui (a, b, c, GMP_RNDN)
#define GS_DIV_UI(a,b,c) mpfr_div_ui (a, b, c, GMP_RNDN)
#define GS_UI_DIV(a,b,c) mpfr_ui_div (a, b, c, GMP_RNDN)
#if (LLL_B == 4)
#define GS_MUL_B(a,b) mpfr_mul_2exp(a,b,2,GMP_RNDN)
#define GS_DIV_B(a,b) mpfr_div_2exp(a,b,2,GMP_RNDN)
#else
#define GS_MUL_B(a,b) mpfr_mul_ui(a,b,LLL_B,GMP_RNDN)
#define GS_DIV_B(a,b) mpfr_div_ui(a,b,LLL_B,GMP_RNDN)
#endif
#define GS_DIV(a,b,c) mpfr_div (a, b, c, GMP_RNDN)
#define GS_LT(a,b) mpfr_cmp (a, b) < 0
#define GS_LT_UI(a,b) mpfr_cmp_ui (a, b) < 0
#define CLEAR mpfr_clear
#define GS_GTABS(a,b) mpfr_cmpabs (a, b) > 0
#define GS_GTABS_ONEHALF(x) ((mpfr_sgn (x) > 0) ? (mpfr_cmp_d (x, 0.5) > 0) \
                                              : (mpfr_cmp_d (x, -0.5) < 0))
#define GS_SWAP mpfr_swap
#define GS_ROUND_DIV(r,a,b) { mpfr_div (r, a, b, GMP_RNDN); \
                            mpfr_div_2exp (r, r, 1, GMP_RNDN); \
                            mpfr_round (r, r); }
/* a <- round(b) = floor ((2b+1)/2) */
#define GS_ROUND(a,b) mpfr_rint (a, b, GMP_RNDN)
#define MPZ_SET_GS wclr_mpz_set_fr
#define GS_OUT(x) mpfr_out_str (stdout, 10, 0, x, GMP_RNDN); printf ("\n")
#elif (TYPE == MPF_TYPE)
#define GSTYPE mpf_t
#define GS_STRING "mpf"
#define GS_HAS_PREC 1
#define GS_SET_DEFAULT_PREC mpf_set_default_prec
#define GS_GET_PREC mpf_get_prec
#define GS_SET_PREC mpf_set_prec
#define INIT(x) mpf_init2 (x, PRECISION)
#define GS_SUB mpf_sub
#define GS_MUL(a,b,c) mpf_mul(a, b, c)
#define GS_SUBMUL(a,b,c,_t) { mpf_mul (_t, b, c); \
                              mpf_sub (a, a, _t); }
#define GS_DIVEXACT(a,b,c) mpf_div (a, b, c)
#define GS_SET(a,b) mpf_set (a, b)
#define GS_SET_Z(a,b) mpf_set_z (a, b)
#define GS_ISZERO(x) mpf_cmp_ui (x, 0) == 0
#define GS_ADDMUL(a,b,c,_t) { mpf_mul (_t, b, c); \
                             mpf_add (a, a, _t); }
#define GS_MUL_UI(a,b,c) mpf_mul_ui (a, b, c)
#define GS_DIV_UI(a,b,c) mpf_div_ui (a, b, c)
#define GS_UI_DIV(a,b,c) mpf_ui_div (a, b, c)
#if (LLL_B == 4)
#define GS_MUL_B(a,b) mpf_mul_2exp(a,b,2)
#define GS_DIV_B(a,b) mpf_div_2exp(a,b,2)
#else
#define GS_MUL_B(a,b) mpf_mul_ui(a,b,LLL_B)
#define GS_DIV_B(a,b) mpf_div_ui(a,b,LLL_B)
#endif
#define GS_DIV mpf_div
#define GS_LT(a,b) mpf_cmp (a, b) < 0
#define GS_LT_UI(a,b) mpf_cmp_ui (a, b) < 0
#define CLEAR mpf_clear
#define GS_GTABS mpf_cmpabs
#define GS_GTABS_ONEHALF(x) ((mpf_sgn (x) > 0) ? (mpf_cmp_d (x, 0.5) > 0) \
                                              : (mpf_cmp_d (x, -0.5) < 0))
#define GS_SWAP mpf_swap
/* round(a/(2b)) = floor((a/b + 1)/2) */
#define GS_ROUND_DIV(r,a,b) { mpf_div (r, a, b); \
                            mpf_add_ui (r, r, 1); \
                            mpf_div_2exp (r, r, 1); \
                            mpf_floor (r, r); }
/* a <- round(b) = floor ((2b+1)/2) */
#define GS_ROUND(a,b) { mpf_mul_2exp (a, b, 1); \
                        mpf_add_ui (a, a, 1); \
                        mpf_div_2exp (a, a, 1); \
                        mpf_floor (a, a); }
#define MPZ_SET_GS mpz_set_f
#define GS_OUT(x) mpf_out_str (stdout, 10, 0, x); printf ("\n")
#else
#error "undefined type"
#endif

#ifndef mpfr_get_exp
#define mpfr_get_exp(x) ((x)->_mpfr_exp)
#endif

typedef struct {
  mpz_t **coeff;
  int NumRows, NumCols;
} mat_ZZ;

void init_lll (void);
unsigned long lllint (mpz_t **, unsigned int, unsigned int);
unsigned long lllint2 (mpz_t **, unsigned int, unsigned int, unsigned int);
void post_copy (mpz_t **, unsigned long, unsigned long);
void Print_maple (mpz_t **, int, int);

#ifdef MEMORY_DEBUG
void *__gmp_default_allocate (size_t);
void *__gmp_default_reallocate (void *, size_t, size_t);
void __gmp_default_free (void *, size_t);
void *tests_allocate (size_t);
void tests_free_nosize (void *);
#endif

